/*
 * GL-Zoomquilt. See http://zoomquilt.nikkki.net/
 */

#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include <gtk/gtkgl.h>
#include <GL/gl.h>

typedef struct _QuiltWindow   QuiltWindow;

#define N_QUILT_IMAGES 46
#define N_QUILT_PARTS 12

struct _QuiltWindow
{
  GtkWidget    *window;
  GtkWidget    *drawing_area;
  GTimer       *timer;
  guint         timeout_id;
  gboolean      visibility;

  gdouble       fps;
  gdouble       time;

  gdouble       speed;
  gdouble       zoom;

  gint          n_rectangles;

  GLuint        textures[N_QUILT_IMAGES * N_QUILT_PARTS];
};


static gboolean
quilt_frame_next (QuiltWindow *qwin)
{
  gdouble time, delta_time;

  time = g_timer_elapsed (qwin->timer, NULL);

  delta_time = time - qwin->time;
  qwin->time = time;

  qwin->zoom += qwin->speed * delta_time;

  if (qwin->visibility)
    gtk_widget_queue_draw (qwin->drawing_area);

  return TRUE;
}


/* draw a frame */
static gboolean
quilt_frame_draw (GtkWidget      *widget,
                  GdkEventExpose *event,
                  gpointer        data)
{
  static gint preload_part = 0;

  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);
  gint j, j0;
  gdouble tf;
  gdouble time1, time2;
  GLfloat color[4] = {1.0, 1.0, 1.0, 1.0};
  QuiltWindow  *qwin = (QuiltWindow *) data;

  time1 = g_timer_elapsed (qwin->timer, NULL);

  /*** OpenGL BEGIN ***/
  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return FALSE;

  glClear (GL_COLOR_BUFFER_BIT);
  glEnable (GL_TEXTURE_2D);

  glMaterialfv (GL_FRONT, GL_AMBIENT_AND_DIFFUSE, color);
  glNormal3f (0, 0, 1.0);

  while (qwin->zoom < 0)
    qwin->zoom += N_QUILT_IMAGES;

  while (qwin->zoom > N_QUILT_IMAGES)
    qwin->zoom -= N_QUILT_IMAGES;

  j0 = floor (qwin->zoom);

  /* use a texture part for the next few set of frames */

  if (qwin->speed > 0.0)
    j = (j0 + qwin->n_rectangles) % N_QUILT_IMAGES;
  else
    j = (j0 + N_QUILT_IMAGES - 1) % N_QUILT_IMAGES;

  glBindTexture (GL_TEXTURE_2D,
                 qwin->textures[j * N_QUILT_PARTS + preload_part]);
  glBegin(GL_POLYGON);
  glTexCoord2f ( 0.0,  0.0);
  glVertex2f   (-0.5, -0.5);
  glTexCoord2f ( 1.0,  0.0);
  glVertex2f   ( 0.5, -0.5);
  glTexCoord2f ( 1.0,  1.0);
  glVertex2f   ( 0.5,  0.5);
  glTexCoord2f ( 0.0,  1.0);
  glVertex2f   (-0.5,  0.5);
  glEnd ();
  preload_part = (preload_part + 1) % N_QUILT_PARTS;

  /* draw the actually visible stuff */
  glPushMatrix ();
  tf = pow (2.0, qwin->zoom - j0);
  glScalef (tf, tf, 1.0);

  for (j = 0; j < qwin->n_rectangles; j++)
    {
      gint i, k, texindex;
      if (qwin->n_rectangles > 1 && j == qwin->n_rectangles - 1)
        color[3] = qwin->zoom - j0;
      else
        color[3] = 1.0;

      glMaterialfv (GL_FRONT, GL_AMBIENT_AND_DIFFUSE, color);
      for (i=0; i < 4; i++)
        for (k = 0; k < 3; k++)
          {
            texindex = (j0 + j) * N_QUILT_PARTS + k*4 + i;
            glBindTexture (GL_TEXTURE_2D,
                           qwin->textures[texindex % (N_QUILT_IMAGES * N_QUILT_PARTS)]);
            glBegin(GL_POLYGON);
            glTexCoord2f ( 0.0,  0.0);
            glVertex2f (-1.0 +     i * 0.5, -1.0 +     k * 2.0 / 3.0);
            glTexCoord2f ( 1.0,  0.0);
            glVertex2f (-1.0 + (i+1) * 0.5, -1.0 +     k * 2.0 / 3.0);
            glTexCoord2f ( 1.0,  1.0);
            glVertex2f (-1.0 + (i+1) * 0.5, -1.0 + (k+1) * 2.0 / 3.0);
            glTexCoord2f ( 0.0,  1.0);
            glVertex2f (-1.0 +     i * 0.5, -1.0 + (k+1) * 2.0 / 3.0);
            glEnd ();
          }

      glScalef (0.5, 0.5, 1.0);
    }

  glPopMatrix ();

  if (gdk_gl_drawable_is_double_buffered (gldrawable))
    gdk_gl_drawable_swap_buffers (gldrawable);
  else
    glFlush ();

  gdk_gl_drawable_gl_end (gldrawable);
  /*** OpenGL END ***/

  time2 = g_timer_elapsed (qwin->timer, NULL);

  return TRUE;
}


/* new window size or exposure */
static gboolean
quilt_configure_event (GtkWidget         *widget,
                       GdkEventConfigure *event,
                       gpointer           data)
{
  QuiltWindow *qwin = (QuiltWindow *) data;
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);
  gdouble texture_scalefactor = 0.5;


  GLfloat h = (GLfloat) (widget->allocation.height) / (GLfloat) (widget->allocation.width) / 0.75;

  /* determine how many rectangles should be shown */
  qwin->n_rectangles = 1;
  while (1024 * texture_scalefactor < widget->allocation.width)
    {
      texture_scalefactor *= 2;
      qwin->n_rectangles += 1;
    }

  /*** OpenGL BEGIN ***/
  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return FALSE;

  glViewport (0, 0, widget->allocation.width, widget->allocation.height);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  if (h < 1.0)
    glOrtho (-1.0, 1.0, 1.0*h, -1.0*h, -1.0, 1.0);
  else
    glOrtho (-1.0/h, 1.0/h, 1.0, -1.0, -1.0, 1.0);

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();

  gdk_gl_drawable_gl_end (gldrawable);
  /*** OpenGL END ***/

  return TRUE;
}


static gboolean
quilt_visibility_update (GtkWidget   *widget,
                         GdkEventAny *event,
                         gpointer     data)
{
  QuiltWindow *qwin = (QuiltWindow *) data;

  switch (event->type)
    {
      case GDK_MAP:
          qwin->visibility = TRUE;
          break;
      case GDK_UNMAP:
          qwin->visibility = FALSE;
          break;
      case GDK_VISIBILITY_NOTIFY:
          qwin->visibility = ((GdkEventVisibility *) event)->state != GDK_VISIBILITY_FULLY_OBSCURED;
          break;
      default:
          g_printerr ("Unknown event in quilt_visibility_update!\n");
          qwin->visibility = TRUE;
          break;
    }

  return TRUE;
}


/* change view angle, exit upon ESC */
static gboolean
quilt_key_press_event (GtkWidget   *widget,
                       GdkEventKey *event,
                       gpointer     data)
{
  static gboolean fullscreen = FALSE;
  static gdouble old_speed = 0.0;
  QuiltWindow *qwin = (QuiltWindow *) data;

  switch (event->keyval)
    {
    case GDK_F11:
    case GDK_F:
    case GDK_f:
      if (fullscreen)
        gtk_window_unfullscreen (GTK_WINDOW (qwin->window));
      else
        gtk_window_fullscreen (GTK_WINDOW (qwin->window));
      fullscreen = !fullscreen;
      break;
    case GDK_Escape:
    case GDK_Q:
    case GDK_q:
      gtk_main_quit ();
      break;
    case GDK_Up:
      qwin->zoom += 0.05;
      if (fabs (qwin->speed) > 0.00001)
        {
          old_speed = qwin->speed;
          qwin->speed = 0.0;
        }
      break;
    case GDK_Down:
      qwin->zoom -= 0.05;
      if (fabs (qwin->speed) > 0.00001)
        {
          old_speed = qwin->speed;
          qwin->speed = 0.0;
        }
      break;
    case GDK_Right:
      qwin->speed += 0.05;
      break;
    case GDK_Left:
      qwin->speed -= 0.05;
      break;
    case GDK_space:
    case GDK_P:
    case GDK_p:
      if (fabs (qwin->speed) > 0.00001)
        {
          old_speed = qwin->speed;
          qwin->speed = 0.0;
        }
      else
        {
          qwin->speed = old_speed;
          old_speed = 0.0;
        }
      break;
    default:
      return FALSE;
    }

  return TRUE;
}


static void
quilt_load_textures (QuiltWindow *qwin)
{
  GdkPixbuf *pixbuf;
  guchar *data, *tex_data, *filename;
  gint i, x, y, y0, rowstride, subindex;
  gint bpp = 3;

  glGenTextures (N_QUILT_IMAGES * N_QUILT_PARTS, qwin->textures);

  g_printerr ("loading textures: ");

  subindex = 0;
  tex_data = g_malloc (256*256*bpp);

  for (i = 0; i < N_QUILT_IMAGES; i++)
    {
      g_printerr (".");
      filename = g_strdup_printf ("new/zq-%02d.jpg", i+1);
      pixbuf = gdk_pixbuf_new_from_file (filename, NULL);

      if (gdk_pixbuf_get_width (pixbuf) != 1024 ||
          gdk_pixbuf_get_height (pixbuf) != 768 ||
          gdk_pixbuf_get_n_channels (pixbuf) != bpp)
        {
          g_printerr ("Texture \"%s\" is invalid. Textures must be\n"
                      "1024x768 pixels and must have %d channels.\n",
                      filename, bpp);
          exit (1);
        }


      data = gdk_pixbuf_get_pixels (pixbuf);
      rowstride = gdk_pixbuf_get_rowstride (pixbuf);

      for (y = 0; y < 3; y++)
        for (x = 0; x < 4; x++)
          {
            for (y0 = 0; y0 < 256; y0++)
              {
                 memcpy (tex_data + y0 * 256 * bpp,
                         data + x * 256 * bpp + (y * 256 + y0) * rowstride,
                         256 * bpp);
              }

            glBindTexture (GL_TEXTURE_2D, qwin->textures[subindex]);

            glPixelStorei (GL_UNPACK_ALIGNMENT, 1);
            glTexImage2D (GL_TEXTURE_2D, 0, 3,
                          256, 256,
                          0, GL_RGB,
                          GL_UNSIGNED_BYTE,
                          tex_data);

            glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
            glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
            glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
            glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
            glShadeModel (GL_FLAT);

            subindex++;
          }

      g_object_unref (pixbuf);
      g_free (filename);
    }

  g_printerr (" done\n");
  g_free (tex_data);
}


static void
quilt_init_gl (GtkWidget *widget,
               gpointer   data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);
  QuiltWindow *qwin = (QuiltWindow *) data;

  static GLfloat pos[4] = {5.0, 5.0, 10.0, 0.0};

  /* create timer */
  if (qwin->timer == NULL)
    qwin->timer = g_timer_new ();

  g_timer_start (qwin->timer);

  /*** OpenGL BEGIN ***/
  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return;

  glLightfv (GL_LIGHT0, GL_POSITION, pos);
  glEnable (GL_LIGHTING);
  glEnable (GL_LIGHT0);

  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable (GL_BLEND);
  glEnable (GL_NORMALIZE);

  quilt_load_textures (qwin);

#if 0
  g_printerr ("\n");
  g_printerr ("GL_RENDERER   = %s\n", (char *) glGetString (GL_RENDERER));
  g_printerr ("GL_VERSION    = %s\n", (char *) glGetString (GL_VERSION));
  g_printerr ("GL_VENDOR     = %s\n", (char *) glGetString (GL_VENDOR));
  g_printerr ("GL_EXTENSIONS = %s\n", (char *) glGetString (GL_EXTENSIONS));
  g_printerr ("\n");
#endif

  gdk_gl_drawable_gl_end (gldrawable);
  /*** OpenGL END ***/
}


QuiltWindow *
quilt_setup_ui ()
{
  GdkGLConfig *glconfig;
  GtkWidget *vbox;
  QuiltWindow *qwin;

  qwin = g_new0 (QuiltWindow, 1);

  qwin->fps = 50;
  qwin->speed = -0.15;
  qwin->zoom = g_random_double_range (0.0, N_QUILT_IMAGES);
  qwin->n_rectangles = 1;

  /*
   * Configure OpenGL-capable visual.
   */

  /* Try double-buffered visual */
  glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGB    |
                                        GDK_GL_MODE_DOUBLE);
  if (glconfig == NULL)
    {
      g_printerr ("*** Cannot find the double-buffered visual.\n");
      g_printerr ("*** Trying single-buffered visual.\n");

      /* Try single-buffered visual */
      glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGB);
      if (glconfig == NULL)
        {
          g_printerr ("*** No appropriate OpenGL-capable visual found.\n");
          exit (1);
        }
    }

  /*
   * Top-level window.
   */

  qwin->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (qwin->window), "Zoomquilt");

  /* Get automatically redrawn if any of their children changed allocation. */
  gtk_container_set_reallocate_redraws (GTK_CONTAINER (qwin->window), TRUE);
  gtk_window_set_default_size (GTK_WINDOW (qwin->window), 800, 600);

  g_signal_connect (G_OBJECT (qwin->window), "delete_event",
                    G_CALLBACK (gtk_main_quit), NULL);

  /*
   * VBox.
   */

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (qwin->window), vbox);
  gtk_widget_show (vbox);

  /*
   * Drawing area for drawing OpenGL scene.
   */

  qwin->drawing_area = gtk_drawing_area_new ();

  /* Set OpenGL-capability to the widget. */
  gtk_widget_set_gl_capability (qwin->drawing_area,
                                glconfig,
                                NULL,
                                TRUE,
                                GDK_GL_RGBA_TYPE);

  gtk_widget_add_events (qwin->drawing_area,
                         GDK_VISIBILITY_NOTIFY_MASK);

  g_signal_connect_after (G_OBJECT (qwin->drawing_area), "realize",
                          G_CALLBACK (quilt_init_gl), qwin);
  g_signal_connect (G_OBJECT (qwin->drawing_area), "configure_event",
                    G_CALLBACK (quilt_configure_event), qwin);
  g_signal_connect (G_OBJECT (qwin->drawing_area), "expose_event",
                    G_CALLBACK (quilt_frame_draw), qwin);

  g_signal_connect (G_OBJECT (qwin->drawing_area), "map_event",
                    G_CALLBACK (quilt_visibility_update), qwin);
  g_signal_connect (G_OBJECT (qwin->drawing_area), "unmap_event",
                    G_CALLBACK (quilt_visibility_update), qwin);
  g_signal_connect (G_OBJECT (qwin->drawing_area), "visibility_notify_event",
                    G_CALLBACK (quilt_visibility_update), qwin);

  g_signal_connect (G_OBJECT (qwin->window), "key_press_event",
                    G_CALLBACK (quilt_key_press_event), qwin);

  gtk_box_pack_start (GTK_BOX (vbox), qwin->drawing_area, TRUE, TRUE, 0);

  gtk_widget_show (qwin->drawing_area);

  /*
   * Show window.
   */

  gtk_widget_show (qwin->window);

  /*
   * Timeout function to maintain framerate
   */
  qwin->timeout_id = g_timeout_add_full (GDK_PRIORITY_REDRAW,
                                         1000 / qwin->fps,
                                         (GSourceFunc) quilt_frame_next,
                                         qwin,
                                         NULL);
  return qwin;
}


int
main (int   argc,
      char *argv[])
{
  QuiltWindow *qwin;

  gtk_init (&argc, &argv);
  gtk_gl_init (&argc, &argv);

  qwin = quilt_setup_ui ();

  gtk_main ();

  return 0;
}

